# dict-mapper

Dict-Mapper library provides a flexible and powerful way to transform the structure of a dictionary based on a given set
of transformation rules. This can be particularly useful when you need to reshape data to fit a specific structure or
format.

## Features

The main function in this module is `map_dict(original:)`, which takes an original dictionary and a specification
dictionary as
inputs, and returns a new dictionary that has been transformed according to the specification.

The specification dictionary defines how the original dictionary should be transformed. Its keys are the keys of the new
dictionary, and its values can be:

- A string representing the key in the original dictionary.
- An instance of the 'skip' class to indicate that the key should be omitted.
- A function that takes the original dictionary and returns a value.

## Quick Start

````python
import uuid
from pit.dict_mapper.mapping import map_dict, calculate_from_original, factory, skip

original = {"first_name": "Mac", "last_name": "Doe", "age": 30, "dob": "1990-01-01"}
specification = {
    "id": factory(lambda: str(uuid.uuid4())),
    "first_name": skip(),
    "last_name": skip(),
    "full_name": calculate_from_original(
        lambda d: d["first_name"] + " " + d["last_name"]
    ),
}

result = map_dict(original, specification, keep_unmapped=True)
print(result)
"""
{
    "age": 30,
    "dob": "1990-01-01",
    "id": "160281d4-a243-46b1-b3ac-23e7de7eb178",
    "full_name": "Mac Doe",
}
"""

````

## Usage

```python
import uuid
from datetime import date, datetime
from pit.dict_mapper.mapping import (
    calculate_from_key,
    calculate_from_original,
    constant,
    date_today,
    datetime_now,
    factory,
    map_dict,
    or_default,
    with_type,
    skip,
)

original = {
    "first_name": "Mac",
    "last_name": "Doe",
    "age": 30,
    "dob": "1990-01-01",
    "address": {
        "street": "127 Manuel Fernando de Soto",
        "suburb": "Ciudad de México",
        "postcode": "07469",
    },
}
specification = {
    "id": factory(lambda: str(uuid.uuid4())),
    "status": constant("ACTIVE"),
    "start_date": or_default("start_date", default=date.today()),
    "created_at": datetime_now(),
    "first_name": skip(),
    "last_name": skip(),
    "full_name": calculate_from_original(
        lambda d: d["first_name"] + " " + d["last_name"]
    ),
    "day_of_birth": with_type("dob", rtype=date),
    "home_address": "address",
    "home_address.suburb": skip(),
    "home_address.country.code": skip(),
}

result = map_dict(original, specification, keep_unmapped=True)

print(result)
"""
{
 "age": 30,
 "dob": "1990-01-01",
 "address": {
             > "street": "127 Manuel Fernando de Soto",
 "suburb": "Ciudad de México",
 "postcode": "07469",
},
 "id": "77e7e267-f3fa-4a94-9f2b-d1e98c135a8c",
 "status": "ACTIVE",
 "start_date": datetime.date(2024, 4, 11),
 "created_at": datetime.datetime(2024, 4, 11, 20, 57, 35, 91853),
 "full_name": "Mac Doe",
 "day_of_birth": datetime.date(1990, 1, 1),
 "home_address": {"street": "127 Manuel Fernando de Soto", "postcode": "07469"},
}
"""
```

In this example, the `map_dict()` function is used to create a new dictionary from the `original` dictionary.
The `specification` dictionary defines how the original dictionary should be transformed.
The `first_name` and `last_name` keys are removed from the result, a new `full_name` key is calculated from the original
dictionary, and the `address` key is renamed to `home_address` in the new dictionary. The `keep_unmapped` parameter is
set to `True`, so all keys in the original dictionary that are not in the specification are kept in the new dictionary.

The `key` names of the `specification` will be used as the keys in the new dictionary. The values of the `specification`
are used to know from where take the value for the new dictionary. The values can be a string representing the key in
the original dictionary, an instance of the `skip` class to indicate that the key should be omitted, or a function that
takes the original dictionary and returns a new value.

The library provides several helper functions that can be used in the specification dictionary:

- `constant(value)`: Assigns the given value to the given specification key.
- `or_default(key:str, default)`: Assigns the value of the key in the dictionary or the default value if the key is not
  present.
- `calculate_from_original(func)`: Applies the given function to the original dictionary and assigns the result to the
  given key.
- `calculate_from_key(key:str, transform, default=None)`: Applies the transform function to the value of the key in the
  dictionary and assigns the result to the given key.
- `factory(func)`: Applies the function with no arguments and assigns the result to the given key. Useful for generating
  unique IDs or initialize objects such as list or dict.
- `datetime_now()`: Assigns the current datetime in the given key.
- `date_today()`: Assigns the current date in the given key.
- `with_type(key:str: str, rtype: Type[Any], default=None, _format: str = None)`: Applies a str to bool,date,datetime
  conversion to the value of the key in the dictionary to the given type.