import copy
import logging
from datetime import datetime, date
from typing import Any, Type

from dictor import dictor

logger = logging.getLogger(__name__)


TRUE_STRINGS = ("yes", "true", "t", "1")
FALSE_STRINGS = ("no", "false", "f", "0")

skip = type("skip", (), {})


def constant(value):
    return lambda _: value


def or_default(key, default):
    return lambda d: dictor(d, key, default=default)


def calculate_from_original(func):
    return lambda d: func(d)


def calculate_from_key(key, transform, default=None):
    return lambda d: transform(dictor(d, key, default=default))


def factory(func):
    return lambda _: func()


def datetime_now():
    return lambda _: datetime.now()


def date_today():
    return lambda _: date.today()


def convert_bool(value: str):
    value_lower = value.lower()
    if value_lower in TRUE_STRINGS:
        return True
    elif value_lower in FALSE_STRINGS:
        return False
    return None


def convert_date(value: str, _format: str) -> date:
    if _format is None:
        return datetime.fromisoformat(value).date()
    else:
        return datetime.strptime(value, _format).date()


def convert_datetime(value: str, _format: str) -> datetime:
    if _format is None:
        return datetime.fromisoformat(value)
    else:
        return datetime.strptime(value, _format)


def with_type(key: str, rtype: Type[Any], default=None, _format: str = None):
    conversion_map = {
        bool: convert_bool,
        date: lambda value: convert_date(value, _format),
        datetime: lambda value: convert_datetime(value, _format),
    }

    def convert(data: dict) -> Any:
        value = dictor(data, key)
        if value is None:
            return default
        elif rtype in conversion_map:
            convert_func = conversion_map[rtype]
            return convert_func(value)
        else:
            return rtype(value)

    return convert


def __set_nested(obj: dict, path: str, value) -> None:
    if "." in path:
        *path, last = path.split(".")
        for bit in path:
            if bit not in obj:
                obj[bit] = {}
            obj = obj[bit]
        obj[last] = value
    else:
        obj[path] = value


def __remove_nested(obj: dict, path: str) -> None:
    if "." in path:
        *path, last = path.split(".")
        key_exist = True
        for bit in path:
            if bit not in obj:
                key_exist = False
                break
            obj = obj[bit]
        if key_exist:
            obj.pop(last, None)
        else:
            logger.warning("Key not found: %s", path)
    else:
        obj.pop(path, None)


def map_dict(
    original_dict: dict,
    specification: dict,
    keep_unmapped: bool = False,
) -> dict:
    """
    Transforms an input dictionary according to a given specification.

    Args:
        original_dict (dict): The original dictionary to be transformed.
        specification (dict): A dictionary specifying how the original dictionary should be transformed.

                              The keys are the keys of the new dictionary and the values can be:
                                - A string representing the key in the original dictionary.
                                - An instance of the 'skip' class to indicate that the key should be omitted.
                                - A function that takes the original dictionary and returns a value. Eg lambda original_dict: original_dict["key"] + "suffix"

                              The following functions are available for use in the specification:
                                - `constant(value)`: Assigns the given value to the given specification key.
                                - `or_default(key, default)`: Assigns the value of the key in the dictionary or the default value if the key is not present.
                                - `calculate_from_original(func)`: Applies the given function to the original dictionary and assigns the result to the given key.
                                - `calculate_from_key(key, transform, default=None)`: Applies the transform function to the value of the key in the dictionary and assigns the result to the given key.
                                - `factory(func)`: Applies the function with no arguments and assigns the result to the given key. Useful for generating unique IDs or initialize objects such as list or dict.
                                - `datetime_now()`: Assigns the current datetime in the given key.
                                - `date_today()`: Assigns the current date in the given key.
                                - `with_type(key: str, rtype: Type[Any], default=None, _format: str = None)`: Applies a str to bool,date,datetime convertion to the value of the key in the dictionary to the given type.
        keep_unmapped (bool, optional): If True, the original dictionary will be kept in the new dictionary.
                                        Defaults to False, which means only keys in specification will be present in the result.

    Returns:
        dict: A new dictionary that has been transformed according to the given specification.
    """
    new_dict = {}
    if keep_unmapped:
        new_dict = copy.deepcopy(original_dict)
    for key_target, key_origin in specification.items():
        if isinstance(key_origin, str):
            value = dictor(original_dict, key_origin)
            __set_nested(new_dict, key_target, copy.deepcopy(value))
        elif isinstance(key_origin, skip):
            __remove_nested(new_dict, key_target)
        elif callable(key_origin):
            __set_nested(new_dict, key_target, key_origin(original_dict))
        else:
            logger.warning(
                "Unexpected type for mapping key_origin:[%s], type:[%s]",
                repr(key_origin),
                type(key_origin),
            )
    return new_dict
