import logging
from datetime import date, datetime

import pytest

from pit.dict_mapper.mapping import (
    calculate_from_key,
    calculate_from_original,
    constant,
    date_today,
    datetime_now,
    factory,
    map_dict,
    or_default,
    with_type,
    skip,
)

logger = logging.getLogger(__name__)


def get_id():
    return "ID"


def calculate_complete_name(original_dict):
    return original_dict["data"]["firstName"] + " " + original_dict["data"]["lastName"]


def generate_status(original_status):
    if original_status is not None:
        return original_status.upper()
    else:
        return "UNDEFINED"


def get_email(emailAddresses):
    email = None
    for emailAddress in emailAddresses:
        if emailAddress["isPrimary"]:
            email = emailAddress["email"]
            break
    return email


@pytest.fixture
def person_response():
    return {
        "data": {
            "id": "f4fe41a5-5da4-480d-8f83-856b8cf5a138",
            "displayName": "Abiran Natanael Salas Hernandez",
            "firstName": "Abiran Natanael",
            "lastName": "Salas Hernandez",
            "dateOfBirth": "1991-10-07",
            "gender": "Hombre",
            "isSupervisor": False,
            "isMarried": "false",
            "employeeNumber": "096-MEX",
            "employmentStatus": "Current Staff",
            "addresses": [
                {
                    "addressType": "Casa",
                    "fullAddress": "Manuel Fernando de Soto 127, Constitución de la República, Mexico City, CDMX",
                    "country": "Mexico",
                    "postcode": "07469",
                    "state": "Ciudad de México",
                    "street": "127 Manuel Fernando de Soto",
                    "suburb": "Ciudad de México",
                    "isPrimary": True,
                    "customFields": {},
                    "emailAddresses": [
                        {
                            "email": "asalas@palo-it.com",
                            "isPrimary": True,
                            "isPersonal": False,
                            "customFields": {},
                        },
                        {
                            "email": "abisaher@gmail.com",
                            "isPrimary": False,
                            "isPersonal": True,
                            "customFields": {},
                        },
                    ],
                }
            ],
            "first_commit": "2022-11-11T22:38:14+00:00",
        }
    }


def test_map(person_response):
    # given:

    specification = {
        "id": factory(get_id),
        "name": "data.displayName",
        "dob": with_type("data.dateOfBirth", rtype=date),
        "gender_str": or_default(key="not_exist", default="Male"),
        "is_supervisor": or_default(key="data.isSupervisor", default=True),
        "complete_name": calculate_from_original(calculate_complete_name),
        "employment_status": calculate_from_key(
            key="data.employmentStatus",
            default="Developer",
            transform=generate_status,
        ),
        "employment": calculate_from_key(
            key="employment", default="Developer", transform=generate_status
        ),
        "dojo": calculate_from_key(key="dojo", transform=generate_status),
        "status": constant("ACTIVE"),
        "first_commit": with_type("data.first_commit", rtype=datetime),
        "created_at": datetime_now(),
        "today": date_today(),
        "isMarried": with_type(key="data.isMarried", rtype=bool),
        "isHaveKids": with_type(key="data.not_exist", rtype=bool, default=False),
        "isTodayBirthday": with_type(key="data.dateOfBirth", rtype=bool),
        "not_exist": "not_exist",
    }

    # when:
    result = map_dict(person_response, specification)

    # then:
    assert result is not None
    assert result["id"] == "ID"
    assert result["name"] == "Abiran Natanael Salas Hernandez"
    assert result["dob"] == datetime.strptime("1991-10-07", "%Y-%m-%d").date()
    assert result["gender_str"] == "Male"
    assert result["is_supervisor"] == False
    assert result["complete_name"] == "Abiran Natanael Salas Hernandez"
    assert result["employment_status"] == "CURRENT STAFF"
    assert result["employment"] == "DEVELOPER"
    assert result["dojo"] == "UNDEFINED"
    assert result["status"] == "ACTIVE"
    assert result["first_commit"] == datetime.fromisoformat("2022-11-11T22:38:14+00:00")
    assert result["created_at"].date() == datetime.today().date()
    assert result["today"] == datetime.today().date()
    assert result["isMarried"] is False
    assert result["isHaveKids"] is False
    assert result["isTodayBirthday"] is None
    assert result["not_exist"] is None


def test_map_no_correct_key_type():
    original = {"first_name": "Mac"}
    specification = {"first_name": 5}

    result = map_dict(original, specification)

    assert len(result) == 0


def test_map_with_keep_unmapped():
    original = {"first_name": "Mac", "last_name": "Doe", "age": 30, "dob": "1990-01-01"}
    specification = {
        "full_name": calculate_from_original(
            lambda d: d["first_name"] + " " + d["last_name"]
        ),
    }

    result = map_dict(original, specification, keep_unmapped=True)

    assert result["first_name"] == "Mac"
    assert result["last_name"] == "Doe"
    assert result["age"] == 30
    assert result["dob"] == "1990-01-01"
    assert result["full_name"] == "Mac Doe"


def test_map_with_keep_unmapped_with_skip():
    original = {
        "first_name": "Mac",
        "last_name": "Doe",
        "age": 30,
        "dob": "1990-01-01",
        "address": {
            "street": "127 Manuel Fernando de Soto",
            "suburb": "Ciudad de México",
            "postcode": "07469",
        },
    }
    specification = {
        "first_name": skip(),
        "last_name": skip(),
        "home_address": "address",
        "home_address.suburb": skip(),
        "home_address.country.code": skip(),
        "full_name": calculate_from_original(
            lambda d: d["first_name"] + " " + d["last_name"]
        ),
    }

    result = map_dict(original, specification, keep_unmapped=True)

    assert result.get("first_name") is None
    assert result.get("last_name") is None
    assert result["age"] == 30
    assert result["dob"] == "1990-01-01"
    assert result["full_name"] == "Mac Doe"
    assert result["full_name"] == "Mac Doe"
    assert result["home_address"] is not None
    assert result["home_address"]["postcode"] == "07469"
    assert result["home_address"]["street"] == "127 Manuel Fernando de Soto"
    assert "suburb" not in result["home_address"]


def test_map_with_nested_specs():
    original = {"first_name": "Mac", "last_name": "Doe", "age": 30, "dob": "1990-01-01"}
    specification = {"name.first_name": "first_name", "name.last_name": "last_name"}

    result = map_dict(original, specification)

    assert result["name"]["first_name"] == "Mac"
    assert result["name"]["last_name"] == "Doe"
    assert "age" not in result
    assert "dob" not in result


def test_with_type_no_correct_type():
    # Preparación de datos y función
    error_converter = with_type(key="dob", rtype=date)
    data = {"dob": "hello"}

    # Ejecución y verificación
    with pytest.raises(Exception):
        error_converter(data)


@pytest.mark.parametrize(
    "val,expected",
    [
        ("true", True),
        ("True", True),
        ("t", True),
        ("1", True),
        ("yes", True),
        ("false", False),
        ("False", False),
        ("f", False),
        ("0", False),
        ("no", False),
    ],
)
def test_with_type_bool(val, expected):
    # Preparación de datos y función
    bool_converter = with_type(key="isSupervisor", rtype=bool)
    data = {"isSupervisor": val}

    # Ejecución y verificación
    assert bool_converter(data) is expected


def test_with_type_int():
    int_converter = with_type(key="age", rtype=int)
    data = {"age": "30"}

    assert int_converter(data) == 30


def test_with_type_float():
    float_converter = with_type(key="rating", rtype=float)
    data = {"rating": "4.5"}

    assert float_converter(data) == 4.5


def test_with_type_date():
    date_converter = with_type(key="dob", rtype=date, _format="%Y-%m-%d")
    data = {"dob": "1990-01-01"}

    assert date_converter(data) == date(1990, 1, 1)


def test_with_type_datetime():
    datetime_converter = with_type(
        key="last_update", rtype=datetime, _format="%Y-%m-%dT%H:%M:%S.%fZ"
    )
    data = {"last_update": "2020-05-20T15:30:00.000Z"}

    assert datetime_converter(data) == datetime(2020, 5, 20, 15, 30)
